package DkZero_Base

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module BuildTrace.MakeSource

Parameters

Signature

Sourcetype semitrace = {
  1. key : Ctx.K.t;
  2. key_kvhash_deps : (Ctx.K.t * string) list;
    (*

    list of key, kvhash

    *)
  3. raw_sz32 : int32;
  4. raw_sz : int;
  5. raw_value : Messages_.Reader.Value.t;
}

A partially loaded trace which has sufficient information for topological sorting but not much else.

Sourceval sort_traces_topologically : t list -> (t list, string) result

sort_traces_topologically traces sorts traces in topological order based on their dependencies. If there is a cycle in the traces, an error is returned.

Sourceval sort_semitraces_topologically : semitrace list -> (semitrace list, string) result

sort_semitraces_topologically semitraces sorts semitraces in topological order based on their dependencies. If there is a cycle in the traces, an error is returned.

Sourceval dehydrate_traces : ?buildlogtrace:unit -> valuestore:MlFront_Core.FilePath.t -> add_constant_exn:(value_id:string -> string -> unit Ctx.Promise.t) -> add_source_file_exn: (source_file_sha256:string -> Ctx.Io.file_object -> bool Ctx.Promise.t) -> t list -> t list Ctx.Promise.t

dehydrate_traces ?buildlogtrace ~valuestore ~add_constant_exn ~add_source_file_exn traces returns traces that can and should be persisted to a constructive trace log.

Any potentially large items are saved in the valuestore directly or saved using add_constant_exn or add_source_file_exn.